{******************************************************************************
                      Pilotlogic Software House
                     
 Package pl_Win_MIDI
 This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)      
                                                                               
   ****** BEGIN LICENSE BLOCK *****     
                                                              
   The contents of this file are used with permission, subject to the Mozilla   
   Public License Version 2.0 (the "License"); you may not use this file except 
   in compliance with the License. You may obtain a copy of the License at      
   https://www.mozilla.org/en-US/MPL/2.0/                                     
                                                                               
   Software distributed under the License is distributed on an "AS IS" basis,   
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for 
   the specific language governing rights and limitations under the License. 
   
   ****** END LICENSE BLOCK *****  
  
**********************************************************************************}
unit cmpKeyboard;

interface

uses Windows, Controls, Classes, Graphics;

type
  //TKeyboardOrientation = (kbHorizontal, kbVertical); 7.1.1 - 7.3
  TKeyboardOrientation = (kbHorizontal_0, kbHorizontal_180, kbVertical_90, kbVertical_270); //new
  // North=0, South=180, West=90, East=270 degrees

  TKeyboardNoteEvent = procedure(Sender: TObject; note, velocity: integer) of object;

  TBaseMidiKeys = class(TGraphicControl)
  protected
    FOrientation: TKeyboardOrientation;
    FnoteWidth: single; //integer;
    FBaseOctave: integer;
    FFirstWhiteKey: integer; //new
    FnWhiteKeys: integer; //new
    FBlackBrush: TBrush;
    FWhiteBrush: TBrush;
    FPen: TPen;
    FVelocityMode: boolean;
    FOnNoteOn: TKeyboardNoteEvent;
    FOnNoteOff: TKeyboardNoteEvent;
    noteMap: set of 0..127;
    currentNote: integer;
    procedure SetnoteWidth(_Width: single); // integer);
    procedure SetBaseOctave(octave: integer);
    procedure SetFirstWhiteKey(keynr: integer); //new
    procedure SetnWhiteKeys(n: integer); //new
    procedure SetOrientation(orient: TKeyboardOrientation);
    procedure SetBlackBrush(Value: TBrush);
    procedure SetWhiteBrush(Value: TBrush);
    procedure SetPen(Value: TPen);
    procedure StyleChanged(Sender: TObject);
    procedure FnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
    procedure FnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
    procedure FnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: integer);
    procedure GetSize(var w, l: integer; var blackw, blackl: single);
    procedure PaintKeyboard; //new
    procedure Paint; override;
    procedure InvertNote(note: integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DecodeNote(note: integer; var octave, key: integer; var sharp: boolean);
    function NoteAt(x, y: integer; var velocity: integer): integer;
    procedure PressNote(note, velocity: integer; GenerateEvent: boolean);
    procedure ReleaseNote(note, aftertouch: integer; GenerateEvent: boolean);
    procedure AllNotesOff(GenerateEvent: boolean);
    property OnNoteOn: TKeyboardNoteEvent read FOnNoteOn write FOnNoteOn;
    property OnNoteOff: TKeyboardNoteEvent read FOnNoteOff write FOnNoteOff;
    procedure _xy(x, y: single; var _x, _y: integer); // new
    procedure _wh(dx, dy: single; var _dx, _dy: integer); // new
    property noteWidth: single {integer} read FnoteWidth write SetnoteWidth;
  published
    property Orientation: TKeyboardOrientation read FOrientation write SetOrientation default kbHorizontal_0;
    // property noteWidth: integer read FnoteWidth write SetnoteWidth default 10; //7.1.1
    property BaseOctave: integer read FBaseOctave write SetBaseOctave default 3;
    property FirstWhiteKey: integer read FFirstWhiteKey write SetFirstWhiteKey default 0;  //new
    property nWhiteKeys: integer read FnWhiteKeys write SetnWhiteKeys default 7;  //new
    property Height default 50;
    property Width default 200;
    property BlackBrush: TBrush read FBlackBrush write SetBlackBrush;
    property WhiteBrush: TBrush read FWhiteBrush write SetWhiteBrush;
    property Pen: TPen read FPen write SetPen;
    property VelocityMode: boolean read FVelocityMode write FVelocityMode default True;
    property Align;
  end;


  TMidiKeys = class(TBaseMidiKeys)
  published
    property OnNoteOn;
    property OnNoteOff;
  end;

  TMIDIChannel = 1..16;

  TMidiPiano = class(TBaseMidiKeys)
  private
    FMIDIDevice: integer;
    FMIDIChannel: TMIDIChannel;
    FMIDIPortName: string;
    FMIDIPortOk: boolean;
    FMIDIPort: integer;
    procedure SetMIDIDevice(Value: integer);
    procedure NoteOn(Sender: TObject; note, velocity: integer);
    procedure NoteOff(Sender: TObject; note, velocity: integer);
  protected
    procedure SelectMIDIPort;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property MIDIPort: integer read FMIDIPort;
  published
    property MIDIDevice: integer read FMIDIDevice write SetMIDIDevice nodefault;
    property MIDIChannel: TMIDIChannel read FMIDIChannel write FMIDIChannel default 1;
    property MIDIPortName: string read FMIDIPortName;
    property MIDIPortOk: boolean read FMIDIPortOk;
  end;

//temporary test with xx,yy
//var xx, yy: integer;

implementation

uses SysUtils, MMSystem;

procedure TBaseMidiKeys._xy(x, y: single; var _x, _y: integer);
begin
  // Coordinates transformation formula
  //fi := fi * pi / 180;
  //_x := round(x * cos(fi) + y * sin(fi));
  //_y := round(-x * sin(fi) + y * cos(fi));

  case Orientation of
    kbHorizontal_0:
    begin
      _x := round(x);
      _y := round(y);
    end;
    kbHorizontal_180:
    begin
      _x := Width - 0 - round(x);
      _y := Height - 0 - round(y);
    end;
    kbVertical_90:
    begin
      _x := round(y);
      _y := Height - 0 - round(x);
    end;
    kbVertical_270:
    begin
      _x := Width - round(y);
      _y := round(x);
    end;
  end;//case
end;

procedure TBaseMidiKeys._wh(dx, dy: single; var _dx, _dy: integer);
var
  h: integer;
begin
  _dx := round(dx);
  _dy := round(dy);
  case Orientation of
    kbHorizontal_180:
    begin
      _dx := -_dx;
      _dy := -_dy;
    end;
    kbVertical_90:
    begin
      h := _dx;
      _dx := _dy;
      _dy := -h + 0;
      // dilemma: +0 is ok for sharp keys
      // but +1 is ok for white keys
    end;
    kbVertical_270:
    begin
      h := _dx;
      _dx := -_dy;
      _dy := h;
    end;
  end;//case
end;

constructor TBaseMidiKeys.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];

  FBlackBrush := TBrush.Create;
  FWhiteBrush := TBrush.Create;
  FPen := TPen.Create;

  FBlackBrush.OnChange := StyleChanged;
  FWhiteBrush.OnChange := StyleChanged;
  FPen.OnChange := StyleChanged;

  Width := 200;
  Height := 50;
  FnoteWidth := 10;
  FOrientation := kbHorizontal_0; //new
  FBaseOctave := 3;
  FFirstWhiteKey := 0; //new
  FVelocityMode := True;
  FnWhiteKeys := 7; // new
  BlackBrush.Color := clBlack;
  WhiteBrush.Color := clWhite;
  Pen.Color := clBlack;

  noteMap := [];
  currentNote := -1;

  OnMouseDown := fnMouseDown;
  OnMouseUp := fnMouseUp;
  OnMouseMove := fnMouseMove;
  OnNoteOn := nil;
  OnNoteOff := nil;
end;

destructor TBaseMidiKeys.Destroy;
begin
  FWhiteBrush.Free;
  FBlackBrush.Free;
  FPen.Free;
  inherited Destroy;
end;

procedure TBaseMidiKeys.DecodeNote(note: integer; var octave, key: integer; var sharp: boolean);
var
  k: integer;
begin
  octave := note div 12;
  k := note mod 12;
  if k >= 5 then
    Inc(k);
  key := k shr 1; // white key
  sharp := odd(k);
end;

procedure TBaseMidiKeys.GetSize(var w, l: integer; var blackw, blackl: single); //changed
begin
  case Orientation of
    kbHorizontal_0, kbHorizontal_180:
    begin
      w := Width;
      l := Height;
    end;
    kbVertical_90, kbVertical_270:
    begin
      w := Height;
      l := Width;
    end;
  end; // case
  noteWidth := w / FnWhiteKeys;
  blackl := (0.65 * l); //real size of  piano key
  blackw := (noteWidth * 7 / 12); //real size of piano key
end;

procedure TBaseMidiKeys.PaintKeyboard;
var
  {p, x,} keyNo, x1, y1, x2, y2, i, j: integer;
  p, x: single;
  w, l: integer;
  BlackLength, BlackWidth: single;
begin
  GetSize(w, l, BlackWidth, BlackLength);

  p := noteWidth;
  with canvas do
    while p <= w do
    begin
      // paint white keys
      _xy(p, 0, x1, y1);
      MoveTo(x1, y1);
      _xy(p, l, x1, y1);
      LineTo(x1, y1);
      p:= p+ noteWidth; //Inc(p, noteWidth);
    end;

  p := noteWidth;

  keyNo := 0; // white key
  keyNo := FFirstWhiteKey;
  i := 0;
  x := 0;
  with canvas do
    while x < (w - noteWidth) do
    begin
      j := i mod 12;
      // paint black keys
      x := (i * (noteWidth * 7 / 12) - keyNo * noteWidth);
      if (x > 0) and (x < (w - noteWidth)) then
        if j in [1, 3, 6, 8, 10] then
        begin
          _xy(x, 0, x1, y1);
          _xy(x + BlackWidth, BlackLength, x2, y2);
          Rectangle(x1, y1, x2, y2);
        end;
      Inc(i);
    end;
end;

procedure TBaseMidiKeys.Paint;
var
  w, h: integer; //new
begin
  w := Width;
  h := Height;
  // 7.1.1
  // using Width, Height within Canvas gives error!
  // => bottom line of keyboard is missing
  with Canvas do
  begin
    Brush := FWhiteBrush;
    Pen := FPen;
    // Rectangle(0, 0, Width-1, Height-1); //old : Canvas.Width|Height => bottom line missing
    Rectangle(0, 0, w - 0, h - 0); //new: OK
    Brush := FBLackBrush;
    PaintKeyboard;
  end;
end;

procedure TBaseMidiKeys.InvertNote(note: integer);
var
  Octave, WKeyType, xWhite, xMidi, xMidiNext, xMidiPrev, KeyN, delta: integer;
  sharp, firstK: boolean;
  w, l,  x1, y1, w1, h1: integer;
  BlackLength, BlackWidth, x : single;

  procedure invertWidePart; // under black key
  begin
    x := (xWhite + 2 - FirstWhiteKey * noteWidth);  //+2 OK for orientation 0 & 270; +1 OK for 90 & 180
    _xy(x, BlackLength + 1, x1, y1);
    _wh(noteWidth - 3, l - BlackLength - 3, w1, h1);
    PatBlt(Canvas.Handle, x1, y1, w1, h1, DSTINVERT);
  end;

begin
  Dec(note, 12 * baseOctave);
  GetSize(w, l, BlackWidth, BlackLength);
  DecodeNote(note, Octave, WKeyType, sharp);
  xWhite := round((WKeyType + Octave * 7) * noteWidth);  // white key: vertical & horizontal
  delta := 0; //w mod noteWidth;
  with Canvas do
  begin
    xMidi := round(note * (noteWidth * 7 / 12)); // x of midi key
    xMidiNext := round((note + 1) * (noteWidth * 7 / 12)); // x of midi+1 key
    xMidiPrev := round((note - 1) * (noteWidth * 7 / 12)); // x of midi-1 key
    x := xMidi + 1 - FirstWhiteKey * noteWidth;
    _xy(x, 1, x1, y1);
    _wh(BlackWidth - 2, BlackLength - 2, w1, h1);
    if sharp then
    begin
      firstK := (((Orientation = kbHorizontal_0) or (Orientation = kbHorizontal_180)) and (x1 < (noteWidth / 2))) or
        (((Orientation = kbVertical_90) or (Orientation = kbVertical_270)) and (y1 < (noteWidth / 2)));
      if not firstK then
        PatBlt(Handle, x1, y1, w1, h1, DSTINVERT);
    end
    else
    begin
      KeyN := WKeyType;
      x := xWhite - FirstWhiteKey * noteWidth;
      if (x < noteWidth) and (WKeyType = FirstWhiteKey) then
        case WKeyType of
          0, 1, 3, 4, 5: KeyN := 0;
          2, 6: KeyN := 10; //first white key = 100% white
        end;//case

      if (x >= (w - delta - noteWidth - 1)) then
        case WKeyType of
          1, 2, 4, 5, 6: KeyN := 6;
          0, 3: KeyN := 10; //last white key  = 100% white
        end;//case

      case KeyN of
        0, 3:  // C F
        begin
          invertWidePart;
          x := xWhite + 2 - FirstWhiteKey * noteWidth;
          _xy(x, 2, x1, y1);
          _wh(xMidiNext - xWhite - 1 - 2, BlackLength - 1, w1, h1);
          PatBlt(Handle, x1, y1, w1, h1, DSTINVERT);
        end;

        1, 4, 5: // D G A
        begin
          invertWidePart;
          x := xMidiPrev + blackWidth + 1 - FirstWhiteKey * noteWidth;
          _xy(x, 2, x1, y1);
          _wh(xMidiNext - (xMidiPrev + blackWidth) - 1 - 1, BlackLength - 1, w1, h1);
          PatBlt(Handle, x1, y1, w1, h1, DSTINVERT);
        end;

        2, 6: // E B
        begin
          invertWidePart;
          x := xMidiPrev + blackWidth + 1 - FirstWhiteKey * noteWidth;
          _xy(x, 2, x1, y1);
          _wh(xWhite + noteWidth - (xMidiPrev + blackWidth) - 1 - 1, BlackLength - 1, w1, h1);
          PatBlt(Handle, x1, y1, w1, h1, DSTINVERT);
        end;

        10: // 100% white key at both ends
        begin
          x := xWhite + 2 - FirstWhiteKey * noteWidth;
          _xy(x, 2, x1, y1);
          _wh(noteWidth - 3, l - 4, w1, h1);
          PatBlt(Handle, x1, y1, w1, h1, DSTINVERT);
        end;

      end; //case
    end; //else
  end; // with
end;

procedure TBaseMidiKeys.SetnoteWidth(_Width: single); // integer);
begin
  if abs(FnoteWidth - _Width) > 1e-4 then
  begin
    FnoteWidth := _Width;
    Invalidate;
  end;
end;

procedure TBaseMidiKeys.SetOrientation(Orient: TKeyboardOrientation);
var
  w, l: integer;
  BlackWidth, BlackLength: single;
begin
  if FOrientation <> Orient then
  begin
    FOrientation := Orient;
    GetSize(w, l, BlackWidth, BlackLength); // new
    Repaint;
  end;
end;

procedure TBaseMidiKeys.SetBaseOctave(octave: integer);
begin
  if octave <> FBaseOctave then
  begin
    FBaseOctave := octave;
    Repaint;
  end;
end;

procedure TBaseMidiKeys.SetFirstWhiteKey(keynr: integer);  //new
begin
  if (keynr <> FFirstWhiteKey) then
  begin
    FFirstWhiteKey := abs(keynr);
    Repaint;
  end;
end;

procedure TBaseMidiKeys.SetnWhiteKeys(n: integer);  //new
var
  w, l: integer;
  BlackWidth, BlackLength: single;
begin
  //if n <> FnWhiteKeys then // MidiPianoKeys issue!
  if n <> 0 then
  begin
    FnWhiteKeys := abs(n);
    GetSize(w, l, BlackWidth, BlackLength);
    Repaint;
  end;
end;

procedure TBaseMidiKeys.SetBlackBrush(Value: TBrush);
begin
  FBlackBrush.Assign(Value);
end;

procedure TBaseMidiKeys.SetWhiteBrush(Value: TBrush);
begin
  FWhiteBrush.Assign(Value);
end;

procedure TBaseMidiKeys.SetPen(Value: TPen);
begin
  FPen.Assign(Value);
end;

procedure TBaseMidiKeys.StyleChanged(Sender: TObject);
begin
  Invalidate;
end;

function TBaseMidiKeys.NoteAt(x, y: integer; var velocity: integer): integer;
var
  nWKey, WKeyType, BKeyNo, Octave, noteVolume, delta: integer;
  sharp: boolean;
  w, l, nb: integer;
  BlackLength, BlackWidth, KeyPos: single;
begin
  //temporary test with global xx, yy
  //xx := x;
  //yy := y;

  GetSize(w, l, BlackWidth, BlackLength);
  delta := 0; //w mod noteWidth;
  case Orientation of
    kbHorizontal_0:
    begin
      KeyPos := x;
      noteVolume := y;
    end;
    kbHorizontal_180:
    begin
      KeyPos := w - 1 - x;
      noteVolume := l - 1 - y;
    end;
    kbVertical_90:
    begin
      KeyPos := w - y - 1;
      noteVolume := x;
    end;
    kbVertical_270:
    begin
      KeyPos := y;
      noteVolume := l - x;
    end;
  end;//case

  if KeyPos >= (w - delta) then // unused last unfull key
  begin
    Result := -1;
    exit;
  end;

  // at ends of keyboard
  sharp := not ((KeyPos < (noteWidth / 2)) or (KeyPos > (w - delta - (noteWidth / 2))));

  KeyPos := KeyPos + FirstWhiteKey * noteWidth; // OK

  // white keys
  nWKey := trunc(KeyPos / noteWidth);
  Octave := nWKey div 7;
  WKeyType := nWKey mod 7;

  // midi keys
  BKeyNo := trunc(KeyPos / (noteWidth * 7 / 12));
  nb := BKeyNo + BaseOctave * 12; // Midi key number
  BKeyNo := BKeyNo mod 12;
  sharp := sharp and (noteVolume < BlackLength) and (BKeyNo in [1, 3, 6, 8, 10]);
  if not sharp then  //convert white key to midi key
  begin
    nb := WKeyType * 2;
    if WKeyType > 2 then
      Dec(nb);
    nb := nb + (Octave + BaseOctave) * 12;
  end
  else
    l := round(BlackLength);

  if FVelocityMode then // audio volume
    velocity := noteVolume * 127 div l
  else
    velocity := 127;

  Result := nb;
end;

procedure TBaseMidiKeys.PressNote(note, velocity: integer; GenerateEvent: boolean);
begin
  if not (note in noteMap) then
  begin
    InvertNote(note);
    noteMap := noteMap + [note];
  end;
  if GenerateEvent and Assigned(FOnNoteOn) then
    OnNoteOn(Self, note, velocity);
end;

procedure TBaseMidiKeys.ReleaseNote(note, aftertouch: integer; GenerateEvent: boolean);
begin
  if note in noteMap then
  begin
    InvertNote(note);
    noteMap := noteMap - [note];
  end;
  if GenerateEvent and Assigned(FOnNoteOff) then
    OnNoteOff(Self, note, 0);
end;

procedure TBaseMidiKeys.AllNotesOff(GenerateEvent: boolean);
var
  i: integer;
begin
  for i := 0 to 127 do
    if i in noteMap then
      ReleaseNote(i, 0, GenerateEvent);
end;

procedure TBaseMidiKeys.FnMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  n, velocity: integer;
begin
  SetCaptureControl(self);
  if Button = mbLeft then
  begin
    n := NoteAt(X, Y, velocity);
    if (n >= 0) and (n < 128) then
    begin
      currentNote := n;
      PressNote(currentNote, velocity, True);
    end;
  end;
end;

procedure TBaseMidiKeys.FnMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  SetCaptureControl(nil);
  ReleaseNote(currentNote, 0, True);
end;

procedure TBaseMidiKeys.FnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: integer);
var
  Note, velocity: integer;
begin
  if (ssLeft in Shift) and PtInRect(Rect(0, 0, Width - 1, Height - 1), Point(x, y)) then
  begin
    note := NoteAt(X, Y, velocity);
    if (note >= 0) and (note < 128) and (note <> CurrentNote) then
    begin
      ReleaseNote(CurrentNote, 0, True);
      CurrentNote := note;
      PressNote(CurrentNote, velocity, True);
    end;
  end;
end;

constructor TMidiPiano.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  OnNoteOn := NoteOn;
  OnNoteOff := NoteOff;

  FMIDIDevice := -1;
  FMIDIPortName := 'Microsoft MIDI Mapper';
  FMIDIChannel := 1;
end;

destructor TMidiPiano.Destroy;
begin
  if midiPort <> 0 then
  begin
    midiOutReset(midiPort);
    midiOutClose(midiPort);
  end;
  inherited Destroy;
end;

procedure TMidiPiano.SelectMIDIPort;
var
  rv, noDevs: word;
begin
  if midiPort <> 0 then
  begin
    midiOutReset(midiPort);
    midiOutClose(midiPort);
    FMIDIPort := 0;
  end;

  FMIDIportOk := False;

  noDevs := midiOutGetNumDevs;
  if FMIDIDevice < noDevs then
  begin
    rv := midiOutOpen(@FMIDIPort, FMIDIDevice, 0, 0, 0);
    if rv = 0 then
      FMIDIPortOk := True;
  end;
end;

procedure TMidiPiano.SetMIDIDevice(Value: integer);
var
  noDevs: integer;
  devCaps: MidiOutCaps;
begin
  if Value <> FMIDIDevice then
  begin
    if midiPort <> 0 then
    begin
      midiOutReset(midiPort);
      midiOutClose(midiPort);
      FMIDIPort := 0;
    end;

    FMIDIDevice := Value;
    noDevs := midiOutGetNumDevs;
    if FMIDIDevice < noDevs then
    begin
      midiOutGetDevCaps(FMIDIDevice, @devCaps, sizeof(devCaps));
      FMIDIPortName := StrPas(devCaps.szPName);
    end
    else
      FMIDIPortName := '';
  end;
end;

procedure TMidiPiano.NoteOn(Sender: TObject; note, velocity: integer);
var
  Data: record
    case boolean of
      True: (b1, b2, b3, b4: byte);
      False: (l: longint);
  end;
begin
  if midiPort = 0 then
    SelectMIDIPort;
  if FMIDIPortOk then
    with Data do
    begin
      b1 := $90 + FMIDIChannel - 1;
      b2 := note;
      b3 := velocity;
      b4 := 0;
      midiOutShortMsg(midiPort, l);
    end;
end;

procedure TMidiPiano.NoteOff(Sender: TObject; note, velocity: integer);
var
  Data: record
    case boolean of
      True: (b1, b2, b3, b4: byte);
      False: (l: longint);
  end;
begin
  if FMIDIPortOk then
    with Data do
    begin
      b1 := $80 + FMIDIChannel - 1;
      b2 := note;
      b3 := velocity;
      b4 := 0;
      midiOutShortMsg(midiPort, l);
    end;
end;

end.
